//
//  Copyright (C) 2010 Alexey Bobkov
//
//  This file is part of Fb2toepub converter.
//
//  Fb2toepub converter is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  Fb2toepub converter is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with Fb2toepub converter.  If not, see <http://www.gnu.org/licenses/>.
//

#if 0

#include "hdr.h"

#include <string.h>
#include <string>
#include "streamconv.h"
#include "scanner.h"

#include "unicode/ucnv.h"


namespace Fb2ToEpub
{

//-----------------------------------------------------------------------
// CONVERTER WRAPPER
//-----------------------------------------------------------------------
class ConvBase : public Object
{
public:
    virtual UErrorCode  Convert(char **target, const char *targetLimit, const char **source, const char *sourceLimit, UBool flush) = 0;
    virtual void        Reset() = 0;
};


//-----------------------------------------------------------------------
// CONVERTER TO UNICODE
//-----------------------------------------------------------------------
class ConverterToUnicode : public ConvBase, Noncopyable
{
    UConverter *uc_;
public:
    ConverterToUnicode(const char* name)
    {
        UErrorCode err = U_ZERO_ERROR;
        uc_ = ucnv_open(name, &err);
        if(U_FAILURE(err))
            Error("ucnv_open error");
    }
    ~ConverterToUnicode()
    {
        ucnv_close(uc_);
    }

    //virtuals
    UErrorCode Convert(char **target, const char *targetLimit, const char **source, const char *sourceLimit, UBool flush)
    {
        UChar *t = reinterpret_cast<UChar*>(*target);
        const UChar *tl = reinterpret_cast<const UChar*>(targetLimit);

        UErrorCode err = U_ZERO_ERROR;
        ::ucnv_toUnicode(uc_, &t, tl, source, sourceLimit, NULL, flush, &err);
        *target = reinterpret_cast<char*>(t);
        return err;
    }
    void Reset()
    {
        ::ucnv_reset(uc_);
    }
};


//-----------------------------------------------------------------------
// CONVERTER FROM UNICODE
//-----------------------------------------------------------------------
class ConverterFromUnicode : public ConvBase, Noncopyable
{
    UConverter *uc_;
public:
    ConverterFromUnicode(const char* name)
    {
        UErrorCode err = U_ZERO_ERROR;
        uc_ = ucnv_open(name, &err);
        if(U_FAILURE(err))
            Error("ucnv_open error");
    }
    ~ConverterFromUnicode()
    {
        ucnv_close(uc_);
    }

    //virtuals
    UErrorCode Convert(char **target, const char *targetLimit, const char **source, const char *sourceLimit, UBool flush)
    {
        const UChar *s = reinterpret_cast<const UChar*>(*source);
        const UChar *sl = reinterpret_cast<const UChar*>(sourceLimit);

        UErrorCode err = U_ZERO_ERROR;
        ::ucnv_fromUnicode(uc_, target, targetLimit, &s, sl, NULL, flush, &err);
        *source = reinterpret_cast<const char*>(s);
        return err;
    }
    void Reset()
    {
        ::ucnv_reset(uc_);
    }
};



//-----------------------------------------------------------------------
// stream converter buffer sizes
const size_t IN_CONVBUF_SIZE = 256;
const size_t OUT_CONVBUF_SIZE = 512;

//-----------------------------------------------------------------------
// InConvStm implementation
//-----------------------------------------------------------------------
class InConvStm : public InStm, Noncopyable
{
    Ptr<InStm>          stm_;                       // input stream
    Ptr<ConvBase>       conv_;                      // converter
    mutable char        ibuf_[IN_CONVBUF_SIZE];     // input buffer
    mutable char        *iend_;                     // input buffer unconverted data end
    mutable char        obuf_[OUT_CONVBUF_SIZE];    // output buffer
    mutable char        *ocur_;                     // output buffer current position
    mutable char        *oend_;                     // output buffer concerted data end

    size_t Fill() const;

public:
    InConvStm(InStm *stm, ConvBase *conv);

    //virtuals
    bool        IsEOF() const;
    char        GetChar();
    size_t      Read(void *buffer, size_t max_cnt);
    void        UngetChar(char c);
    void        Rewind();
};

//-----------------------------------------------------------------------
InConvStm::InConvStm(InStm *stm, ConvBase *conv)
                            :   stm_(stm),
                                conv_(conv),
                                iend_(ibuf_),
                                ocur_(obuf_),
                                oend_(obuf_)
{
}

//-----------------------------------------------------------------------
bool InConvStm::IsEOF() const
{
    return (ocur_ == oend_) && (stm_->IsEOF() || !Fill());
}

//-----------------------------------------------------------------------
size_t InConvStm::Fill() const
{
    // fix output pointers
    ocur_ = oend_ = obuf_;

    // read data to input buffer
    iend_ += stm_->Read(iend_, ibuf_ + sizeof(ibuf_) - iend_);
    bool eof = stm_->IsEOF();
    if(iend_ == ibuf_ && !eof)
        Error("InConvStm: internal error");

    // convert data
    const char  *pi = ibuf_;
    UErrorCode err = conv_->Convert(&oend_, oend_ + sizeof(obuf_), &pi, iend_, eof);
    if(U_FAILURE(err) && err != U_BUFFER_OVERFLOW_ERROR)
        Error("ucnv: conversion error");

    // fix input data and pointers
    size_t inleft  = iend_ - pi;
    iend_ = ibuf_ + inleft;
    if(inleft)
        ::memmove(ibuf_, pi, inleft);   // move unconverted rest to beginning

    return oend_ - ocur_;
}

//-----------------------------------------------------------------------
char InConvStm::GetChar()
{
    if(ocur_ == oend_ && !Fill())
        Error("conv: EOF");
    return *ocur_++;
}

//-----------------------------------------------------------------------
size_t InConvStm::Read(void *buffer, size_t max_cnt)
{
    char *pc = reinterpret_cast<char*> (buffer);
    for (size_t cnt = 0; cnt < max_cnt;)
    {
        size_t num_to_copy = oend_ - ocur_;
        if (num_to_copy <= 0 && (num_to_copy = Fill ()) == 0)
            return cnt;
        if (num_to_copy > max_cnt - cnt)
            num_to_copy = max_cnt - cnt;

        ::memcpy (pc, ocur_, num_to_copy);

        pc      += num_to_copy;
        ocur_   += num_to_copy;
        cnt     += num_to_copy;
    }
    return max_cnt;
}

//-----------------------------------------------------------------------
void InConvStm::UngetChar(char c)
{
    if(ocur_ == obuf_)
        Error("conv: can't unget");
    --ocur_;
}

//-----------------------------------------------------------------------
void InConvStm::Rewind()
{
    stm_->Rewind();
    conv_->Reset();
    iend_ = ibuf_;
    ocur_ = oend_ = obuf_;
}


//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
Ptr<InStm> CreateInConvStm(InStm *stm, const char* tocode, const char* fromcode)
{
    Ptr<ConvBase> conv1 = new ConverterToUnicode(fromcode);
    Ptr<ConvBase> conv2 = new ConverterFromUnicode(tocode);
    Ptr<InStm> stm1 = new InConvStm(stm, conv1);
    return new InConvStm(stm1, conv2);
}


//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static const char* RoughEncoding(InStm *stm)
{
    // usually it's enough to test the first byte only...
    switch(stm->GetUChar())
    {
    case 0x00:  return "UTF-32BE";
    case 0x0E:  return "SCSU";
    case 0x2B:  return "UTF-7";
    case '<':   return "UTF-8";     // temporary assume UTF-8 XML
    case 0x84:  return "GB-18030";
    case 0xEF:  return "UTF-8";
    case 0xDD:  return "UTF-EBCDIC";
    case 0xF7:  return "UTF-1";
    case 0xFB:  return "BOCU-1";
    case 0xFE:  return "UTF-16BE";
    case 0xFF:
        if(stm->GetUChar() != 0xFE)
            break;
        return stm->GetUChar() ? "UTF-16LE" : "UTF-32LE";
    default:
        break;
    }

    Error("bad XML or unknown encoding");
    return NULL;
}

//-----------------------------------------------------------------------
// INPUT UTF-8 STREAM (DISCARD BYTE ORDER MARK)
//-----------------------------------------------------------------------
class InStmUtf8 : public InStm, Noncopyable
{
    Ptr<InStm>  stm_;
    bool        has_bom_;

public:
    InStmUtf8(InStm *stm, const char *fromcode);

    //virtuals
    bool        IsEOF() const                       {return stm_->IsEOF();}
    char        GetChar()                           {return stm_->GetChar();}
    size_t      Read(void *buffer, size_t max_cnt)  {return stm_->Read(buffer, max_cnt);}
    void        UngetChar(char c)                   {return stm_->UngetChar(c);}
    void        Rewind();
};

//-----------------------------------------------------------------------
InStmUtf8::InStmUtf8(InStm *stm, const char *fromcode)
{
    stm->Rewind();
    stm_ =  strcmp(fromcode, "UTF-8") ? CreateInConvStm(stm, "UTF-8", fromcode) : stm;
    unsigned char uc = stm_->GetUChar();
    // skip byte order mark
    has_bom_ = (uc == 0xEF);
    if(!has_bom_)
        stm_->UngetUChar(uc);
    else if(stm_->GetUChar() != 0xBB || stm_->GetUChar() != 0xBF)
        Error("internal: bad UTF-8 BOM");
}

//-----------------------------------------------------------------------
void InStmUtf8::Rewind()
{
    stm_->Rewind();
    if(has_bom_)
    {
        // skip byte order mark
        stm_->GetChar();
        stm_->GetChar();
        stm_->GetChar();
    }
}

static std::string ParseEncoding(InStm *stm)
{
    Ptr<LexScanner> scanner = CreateScanner(stm);

    // skip version
    if (scanner->GetToken() != LexScanner::Token(LexScanner::XMLDECL) ||
        scanner->GetToken() != LexScanner::Token('"') ||
        scanner->GetToken().type_ != LexScanner::VERSION ||
        scanner->GetToken() != LexScanner::Token('"'))
    {
        Error("encoding parsing error 1");
    }

    // check if next is encoding
    LexScanner::Token t = scanner->GetToken();
    if(t != LexScanner::Token(LexScanner::ENCODING))
        return "UTF-8";

    // skip '"'
    if (scanner->GetToken() != LexScanner::Token('"'))
        Error("encoding parsing error 2");

    // next should be string value
    t = scanner->GetToken();
    if(t.type_ != LexScanner::VALUE)
        Error("encoding parsing error 3");
    return t.s_;
}

//-----------------------------------------------------------------------
Ptr<InStm> FB2TOEPUB_DECL CreateInUnicodeStm(InStm *stm)
{
    // analize bit order mark to get rough encoding
    const char *fromcode = RoughEncoding(stm);

    // parse <?xml version="1.0" encoding="..."?>
    // to refine the encoding
    Ptr<InStm> tmp = new InStmUtf8(stm, fromcode);
    return new InStmUtf8(stm, ParseEncoding(tmp).c_str());
}


};  //namespace Fb2ToEpub

#endif
